@@ -1,4 +1,4 @@
-import os, string, random, json, sys, subprocess, fcntl, time, traceback, socket, urllib.request, urllib.error, tarfile
+import os, string, random, json, sys, subprocess, fcntl, time, traceback, socket, urllib.request, urllib.error, tarfile, pg8000, pg8000.dbapi, glob
 
 
 def connect(dirname="data/postgresqlite", sqlite_compatible=True, config=None):
@@ -14,9 +14,7 @@ def connect(dirname="data/postgresqlite", sqlite_compatible=True, config=None):
         config (Config | None): An object obtained through `get_config()` can be given
             to configure the connection. This causes `dirname` to be ignored.
     """
-    import pg8000.dbapi
-
-    if sqlite_compatible:
+    if sqlite_compatible and pg8000.dbapi.Cursor.fetchall != _cursor_fetchall:
         pg8000.dbapi.paramstyle = "qmark"
 
         pg8000.dbapi.Connection.execute = _conn_execute
@@ -26,8 +24,22 @@ def connect(dirname="data/postgresqlite", sqlite_compatible=True, config=None):
         pg8000.dbapi.Cursor.__iter__ = _cursor_iter
 
     config = config or get_config(dirname)
-    connection = pg8000.dbapi.connect(user=config.user, password=config.password, unix_sock=config.socket)
+
+    retries = 0
+    while True:
+        try:
+            connection = pg8000.dbapi.connect(user=config.user, password=config.password, unix_sock=config.socket)
+            connection.cursor().execute('SELECT 1')
+        except Exception as e:
+            if 'the database system is starting up' not in str(e) or retries >= 50:
+                raise e
+            retries += 1
+            time.sleep(0.2)
+            continue
+        break
+        
     if sqlite_compatible:
+        connection.cursor().execute('COMMIT')
         connection.autocommit = True
     return connection
 
@@ -171,7 +183,7 @@ def _auto_start(config):
     else:
         _download_server(config)
 
-        log_fd = open(config.dir+"/postgresqlite.log", "w")
+        log_fd = open(config.dir+"/postgresqlite.log", "a", buffering=1)
 
         if not os.path.exists(config.dir+"/pgdata/postgresql.conf"):
             print("Initializing new PostgreSQL data dir..", file=sys.stderr)
@@ -189,7 +201,7 @@ def _auto_start(config):
             os.remove(password_file)
 
         print("Starting PostgreSQL..", file=sys.stderr)
-        _run_as_daemon(lambda: _run_server(daemon_fd, log_fd, config))
+        _run_as_daemon(lambda: _run_server(daemon_fd, log_fd, config), keep_fds={daemon_fd,log_fd}, change_cwd=False)
 
     client_file = lockdir + "/" + _make_random_word()
     global client_fd # to make sure the GC doesn't close our file
@@ -278,6 +290,8 @@ class Config:
 
 
 def _run_server(daemon_fd, log_fd, config):
+    sys.stdout = sys.stderr = log_fd
+    print("PostgreSQL daemon is starting..")
 
     lockdir = config.dir+"/locks"
     proc = None
@@ -346,51 +360,63 @@ def _run_server(daemon_fd, log_fd, config):
         pass
     daemon_fd.close()
 
-    if proc:
+    if proc != None:
         proc.terminate()
         try:
             proc.wait(10)
-            print(f"PostgreSQLite shutdown successfull", file=log_fd, flush=True)
+            print(f"PostgreSQLite shutdown successful", file=log_fd, flush=True)
         except subprocess.TimeoutExpired:
             print(f"PostgreSQLite killing server", file=log_fd, flush=True)
             proc.kill()
             proc.wait()
 
     try:
+        for filename in glob.glob(config.socket_dir + "/.s.PGSQL.*"):
+            os.unlink(filename)
         os.rmdir(config.socket_dir)
-    except:
-        print(f"Couldn't delete {config.socket_dir}", file=log_fd, flush=True)
+    except Exception as e:
+        print(f"Couldn't delete {config.socket_dir}", e, file=log_fd, flush=True)
 
     log_fd.close()
 
 
-def _run_as_daemon(daemon_callback):
+def _run_as_daemon(daemon_callback, keep_fds=set(), change_cwd=True):
     # Do the Unix double-fork magic; see Stevens's book "Advanced
     # Programming in the UNIX Environment" (Addison-Wesley) for details
-    try:
-        pid = os.fork()
-        if pid > 0:
-            # Return the original process
-            return
-    except OSError as e:
-        print(f"fork #1 failed: {e.errno} ({e.strerror})", file=sys.stderr)
-        sys.exit(1)
+    pid = os.fork()
+    if pid > 0:
+        # Return the original process
+        return
 
     # Decouple from parent environment
-    os.chdir("/")
-    os.setsid(  )
+    if change_cwd:
+        os.chdir("/")
+    os.setsid()
     os.umask(0)
 
+    keep_fds = {fd if type(fd)==int else fd.fileno() for fd in keep_fds}
+    for file in [sys.stdout, sys.stderr, sys.stdin]:
+        try:
+            if file.fileno() not in keep_fds:
+                file.close()
+        except:
+            pass
+    for fd in range(1024):
+        if fd not in keep_fds:
+            try:
+                os.close(fd)
+            except:
+                pass
+
     # Do second fork
+    pid = os.fork()
+    if pid > 0:
+        # Exit from second parent; print eventual PID before exiting
+        sys.exit(0)
+
     try:
-        pid = os.fork()
-        if pid > 0:
-            # Exit from second parent; print eventual PID before exiting
-            sys.exit(0)
-    except OSError as e:
-        print(f"fork #2 failed: {e.errno} (e.strerror)", file=sys.stderr)
-        sys.exit(1)
-
-    daemon_callback()
+        daemon_callback()
+    except Exception as e:
+        # Make sure an exception is not caught by _run_as_daemon's caller
+        print(traceback.format_exc(), file=sys.stderr, flush=True)
     sys.exit(0)
-
